==> orcalib-0.0.79.tar.gz: orcalib-0.0.79/orcalib/analysis/analysis_duplicate.py <==
import hashlib

from PIL import Image as pil
from pydantic import BaseModel

from ..memoryset import LabeledMemory, LabeledMemoryLookup
from ..pydantic_utils import UUID7, InputType
from .analysis import MemorysetAnalysis


class MemoryDuplicateMetrics(BaseModel):
    is_duplicate: bool
    duplicate_memory_ids: list[UUID7]
    has_possible_duplicates: bool
    possible_duplicate_memory_ids: list[UUID7]


class MemorysetDuplicateMetrics(BaseModel):
    num_duplicates: int
    """The number of duplicate memories in the memoryset"""


class MemorysetDuplicateAnalysisConfig(BaseModel):
    possible_duplicate_threshold: float = 0.97
    """Lookup score threshold for a memory to be considered a possible duplicate"""


class MemorysetDuplicateAnalysis(
    MemorysetAnalysis[MemorysetDuplicateAnalysisConfig, MemoryDuplicateMetrics, MemorysetDuplicateMetrics]
):
    """
    Find exact and possible duplicates in a memoryset

    Note:
        This does not remove duplicates since that would not be compatible with running this
        alongside other analysis that depend on memoryset consistency while the analysis runs.
        To delete duplicates after the analysis has finished, you can, for example, run:

        ```py
        memoryset.delete(
            m.memory_id
            for m in memoryset.query(
                filters=[("metrics.is_duplicate", "==", True)]
            )
        )
        ```
    """

    name = "duplicate"

    def __init__(self, config: MemorysetDuplicateAnalysisConfig | None = None, **kwargs):
        self.config = config or MemorysetDuplicateAnalysisConfig(**kwargs)

        self._duplicate_memory_ids: set[UUID7] = set()

    @staticmethod
    def is_duplicate_value(v1: InputType, v2: InputType) -> bool:
        if isinstance(v1, pil.Image) and isinstance(v2, pil.Image):
            hash_image = lambda image: hashlib.md5(image.convert("RGB").tobytes()).hexdigest()
            return hash_image(v1) == hash_image(v2)
        else:
            return v1 == v2

    def on_batch(
        self, memories_batch: list[LabeledMemory], neighbors_batch: list[list[LabeledMemoryLookup]]
    ) -> list[tuple[UUID7, MemoryDuplicateMetrics]]:
        metrics: list[tuple[UUID7, MemoryDuplicateMetrics]] = []
        for i, memory in enumerate(memories_batch):
            duplicate_memory_ids: list[UUID7] = []
            possible_duplicate_memory_ids: list[UUID7] = []
            for neighbor in neighbors_batch[i]:
                if neighbor.memory_id in self._duplicate_memory_ids:
                    continue
                is_duplicate = self.is_duplicate_value(memory.value, neighbor.value)
                if is_duplicate:
                    duplicate_memory_ids.append(neighbor.memory_id)
                    self._duplicate_memory_ids.add(memory.memory_id)
                elif neighbor.lookup_score > self.config.possible_duplicate_threshold:
                    possible_duplicate_memory_ids.append(neighbor.memory_id)

            metrics.append(
                (
                    memory.memory_id,
                    MemoryDuplicateMetrics(
                        is_duplicate=len(duplicate_memory_ids) > 0,
                        duplicate_memory_ids=duplicate_memory_ids,
                        has_possible_duplicates=len(possible_duplicate_memory_ids) > 0,
                        possible_duplicate_memory_ids=possible_duplicate_memory_ids,
                    ),
                )
            )

        return metrics

    def after_all(self) -> MemorysetDuplicateMetrics:
        return MemorysetDuplicateMetrics(
            num_duplicates=len(self._duplicate_memory_ids),
        )

==> orcalib-0.0.88.tar.gz: orcalib-0.0.88/orcalib/analysis/analysis_duplicate.py <==
import hashlib

import numpy as np
from PIL import Image as pil
from pydantic import BaseModel

from ..memoryset import LabeledMemory, LabeledMemoryLookup
from ..utils.pydantic import UUID7, InputType
from .analysis import MemorysetAnalysis


class MemoryDuplicateMetrics(BaseModel):
    is_duplicate: bool
    duplicate_memory_ids: list[UUID7]
    has_potential_duplicates: bool
    potential_duplicate_memory_ids: list[UUID7]


class MemorysetDuplicateMetrics(BaseModel):
    num_duplicates: int
    """The number of duplicate memories in the memoryset"""


class MemorysetDuplicateAnalysisConfig(BaseModel):
    potential_duplicate_threshold: float = 0.97
    """Lookup score threshold for a memory to be considered a potential duplicate"""


class MemorysetDuplicateAnalysis(
    MemorysetAnalysis[MemorysetDuplicateAnalysisConfig, MemoryDuplicateMetrics, MemorysetDuplicateMetrics]
):
    """
    Find exact and potential duplicates in a memoryset

    Note:
        This does not remove duplicates since that would not be compatible with running this
        alongside other analysis that depend on memoryset consistency while the analysis runs.
        To delete duplicates after the analysis has finished, you can, for example, run:

        ```py
        memoryset.delete(
            m.memory_id
            for m in memoryset.query(
                filters=[("metrics.is_duplicate", "==", True)]
            )
        )
        ```
    """

    name = "duplicate"

    def __init__(self, config: MemorysetDuplicateAnalysisConfig | None = None, **kwargs):
        self.config = config or MemorysetDuplicateAnalysisConfig(**kwargs)

        self._duplicate_memory_ids: set[UUID7] = set()

    @staticmethod
    def is_duplicate_value(v1: InputType, v2: InputType) -> bool:
        if isinstance(v1, pil.Image) and isinstance(v2, pil.Image):
            hash_image = lambda image: hashlib.md5(image.convert("RGB").tobytes()).hexdigest()  # noqa: E731
            return hash_image(v1) == hash_image(v2)
        elif isinstance(v1, np.ndarray) and isinstance(v2, np.ndarray):
            return np.allclose(v1, v2)
        else:
            return v1 == v2

    def on_batch(
        self, memories_batch: list[LabeledMemory], neighbors_batch: list[list[LabeledMemoryLookup]]
    ) -> list[tuple[UUID7, MemoryDuplicateMetrics]]:
        metrics: list[tuple[UUID7, MemoryDuplicateMetrics]] = []
        for i, memory in enumerate(memories_batch):
            duplicate_memory_ids: list[UUID7] = []
            potential_duplicate_memory_ids: list[UUID7] = []
            for neighbor in neighbors_batch[i]:
                if neighbor.memory_id in self._duplicate_memory_ids:
                    continue
                is_duplicate = self.is_duplicate_value(memory.value, neighbor.value)
                if is_duplicate:
                    duplicate_memory_ids.append(neighbor.memory_id)
                    self._duplicate_memory_ids.add(memory.memory_id)
                elif neighbor.lookup_score > self.config.potential_duplicate_threshold:
                    potential_duplicate_memory_ids.append(neighbor.memory_id)

            metrics.append(
                (
                    memory.memory_id,
                    MemoryDuplicateMetrics(
                        is_duplicate=len(duplicate_memory_ids) > 0,
                        duplicate_memory_ids=duplicate_memory_ids,
                        has_potential_duplicates=len(potential_duplicate_memory_ids) > 0,
                        potential_duplicate_memory_ids=potential_duplicate_memory_ids,
                    ),
                )
            )

        return metrics

    def after_all(self) -> MemorysetDuplicateMetrics:
        return MemorysetDuplicateMetrics(
            num_duplicates=len(self._duplicate_memory_ids),
        )
